# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
"""
 this file provides the Zhang and Shasha tree to tree correction algorithm
 extended by Barnard, Clark and Duncan  
"""

__revision__ = "$Id: ezs.py,v 1.8 2004/09/03 10:06:38 alf Exp $"

from logilab.xmldiff.objects import *
from logilab.xmldiff.misc import init_matrix

####### Actions used by ezs algorithm #######
EZS_A_TYPE  = 0
EZS_A_COST  = 1
EZS_A_FCOST = 2
EZS_A_N1    = 3
EZS_A_N2    = 4
EZS_A_DEL   = 5
# node's attributes for ezs algorithm
N_KEYROOT  = NSIZE # 1 if node is a keyroot, either 0 
N_LEFTMOST = N_KEYROOT+1 # index of leftmost child (see tree2tree)

def _nodes_equal(n1, n2):
    """ compare name and value of 2 xml nodes n1 and n2 """
    if n1 is None:
        return n2 is None
    elif n2 is None:
        return FALSE    
    return n1[N_VALUE] == n2[N_VALUE] 

def trees_equal(n1, n2):
    """ return true if the node n1 and n2 are equivalent subtrees """
    if not _nodes_equal(n1, n2):
        return FALSE
    elif n1 is not None:
        # recursion on each n1 and n2's child
        if n1[N_ISSUE] != n2[N_ISSUE]:
            return FALSE
        else:
            for child1, child2 in zip(n1[N_CHILDS], n2[N_CHILDS]):
                if not trees_equal(child1, child2):
                    return FALSE
    return TRUE

def choose(f_actions, desc_list):
    """
    return the best action (min forest distance) in the description list
    desc_list : [index1, index2, Action]
    """
    best_action = [C_INFINI]
    for i, j, action in desc_list:
        fcost = f_actions[i][j][-1][EZS_A_FCOST] + action[EZS_A_COST]
        if fcost < best_action[0]:
            best_action = [fcost, i, j, action]
    actions_stack = f_actions[best_action[1]][best_action[2]][:]
    best_action[3][EZS_A_FCOST]  = best_action[0]
    add_action(actions_stack, best_action[3])
    return actions_stack

def add_action(actions_list, action):
    """ Test action and add it to the list if it's a real action """
    if action[EZS_A_COST] > 0:
        actions_list.append(action)

######## COST CALCUL ########
C_INFINI = 9999999
C_SWAP   = 1
C_APPEND = 1
C_REMOVE = 1

def gamma(ni, nj):
    """
    return a cost which represents the differents betwen ni and nj
    today, return 0 if ni.nodeName equal nj.nodeName, 1 else
    """
    if ni == nj :
        return 0
    elif ni is not None and nj is not None and ni[N_VALUE] == nj[N_VALUE]:
        return 0
    else:
        return C_INFINI

def swap_trees(ni, sib_o_fi, nj, sib_o_fj):
    """
    return the cost to swap subtree ni et nj
    (sib_o_fi and sib_o_fi are the next sibbling node
     respectively for n1 and nj)
    """
    if trees_equal(ni, sib_o_fj) and trees_equal(sib_o_fi, nj) \
       and ni[N_NAME] != '/' and nj[N_NAME] != '/':
        return C_SWAP
    else:
        return C_INFINI    

def remove_tree(ni):
    """ return the cost to remove subtree ni """
    return (ni[N_ISSUE] + 1) * C_REMOVE

def append_tree(ni):
    """ return the cost to append subtree ni """
    return (ni[N_ISSUE] + 1) * C_APPEND

##### TREE 2 TREE ALGORITHMs ###
class EzsCorrector:
    """
    this class uses the Zhang and Shasha algorithm extended by Barnard, Clark 
    and Duncan 
    """
## * x, y                        -> postordered number of nodes being processed
## * nl1, nl2: node[MAXNODES]    -> nodes list ordered in the post-ordered 
##    number extracted respectively from tree1 and tree2 (size1 and size2 elmts)
## * actions[MAXNODES][MAXNODES] -> actions table working as tree distances
##     table (consideres only descendants). actions[i][j] finally contain a
##     list of actions which represents the best way to transform node
##     post-numbered i (from source tree) into node post-numbered j (from
##     destination tree)
## * f_actions[FDSIZE][FDSIZE]   -> actions table working as forest distances
##     table (forest distance is the distance between 2 nodes in their left
##     siblings context)
## since nodes are post numbered, nl[nl[i]->leftmost]-1] = root of the previous
## subtree for nl[i]
    
    def __init__(self, formatter):
        self._formatter = formatter

    def process_trees(self, tree1, tree2):
        """
        the Extended Zhang and Shasha tree 2 tree correction algorithm
        (Barnard, Clarke, Duncan)
        """
        ### initialisations ###
        nl1, nl2 = [], []
        self._formatter.init()
        # add attributes to trees
        self._post_order(tree1, nl1, TRUE)
        self._post_order(tree2, nl2, TRUE)
        # numbered tree with required attributes
        size1, size2 = len(nl1), len(nl2)
        # actions tables init
        f_actions = init_matrix(size1+1, size2+1, [[0, 0, C_INFINI, None]])
        actions = init_matrix(size1+1, size2+1, None)
        # insert None elmt to have index from 1 to size instead of 0,size-1
        nl1.insert(0, None)
        nl2.insert(0, None)
        
        ## after that, let's go !! ###
        for x in range(1, size1+1):
            if nl1[x][N_KEYROOT]:
                for y in range(1, size2+1):
                    if nl2[y][N_KEYROOT]:
                        # all the job is in function below
                        self._process_nodes(x, y, nl1, nl2, f_actions, actions) 
                    
        self._mainformat(actions[size1][size2])
        self._formatter.end()

    #### private functions ####
    def _process_nodes(self, x, y, nl1, nl2, f_actions, actions):
        """
        job for each keyroot nodes
        after round for nodes (nl1[x], nl2[y]), actions[x][y] will contain 
        the best list of actions to transform nl1[x] into nl2[y]
        (f_actions[x][y] too but it may be override in the next round)
        """
        lx = nl1[x][N_LEFTMOST]
        ly = nl2[y][N_LEFTMOST]
        f_actions[lx - 1][ly - 1] = [[0, 0, 0, None]]
        
        # init forrest distance array by the cost of removing and appending
        # each subtree on a cumulative basis
        for i in range(lx, x+1):
            f_actions[i][ly - 1] = f_actions[nl1[i][N_LEFTMOST] - 1][ly - 1][:]
            cost = remove_tree(nl1[i])
            add_action(f_actions[i][ly - 1],
                       [AT_REMOVE, cost,
                        f_actions[i][ly - 1][-1][EZS_A_FCOST]+cost,
                        nl1[i]
                        ])

        for j in range(ly, y+1):
            f_actions[lx - 1][j] = f_actions[lx - 1][nl2[j][N_LEFTMOST] - 1][:]
            cost = append_tree(nl2[j])
            add_action(f_actions[lx - 1][j],
                       [AT_APPEND, cost,
                        f_actions[lx - 1][j][-1][EZS_A_FCOST]+cost,
                        nl2[j], nl1[x]
                        ])
    
        # look for the shortest way
        for i in range(lx, x+1):
            for j in range(ly, y+1):
                li = nl1[i][N_LEFTMOST]
                lj = nl2[j][N_LEFTMOST]

                # min cost between gamma, remove(nl1[i]), append(nl2[j], nl1[i])
                f_actions[i][j] = choose(f_actions,
                 [
                 [i-1, j, [AT_REMOVE, gamma(nl1[i], None), 0, nl1[i]]],
                 [i, j-1, [AT_APPEND, gamma(None, nl2[j]), 0, nl2[j], nl1[x]]],
                 [li-1, j, [AT_REMOVE, remove_tree(nl1[i]), 0, nl1[i]]],
                 [i, lj-1, [AT_APPEND, append_tree(nl2[j]), 0, nl2[j], nl1[x]]]
                 ])

                if li == lx and lj == ly:
                    # min between just calculed and last loop + change
                    f_actions[i][j] = choose(f_actions,
                     [
                      [i, j, [0, 0, 0]],
                      [i-1, j-1, [0, gamma(nl1[i], nl2[j]), 0, nl1[i], nl2[j]]]
                     ])
                    # now we got the best way from nl1[i] to nl2[j, save it
                    actions[i][j] = f_actions[i][j][:]
                
                else:
                    if nl1[i][N_KEYROOT] and nl2[j][N_KEYROOT] \
                       and nl1[i][N_NAME] != '/' and nl2[j][N_NAME] != '/':
                        # min between just calculed and swap
                        f_actions[i][j] = choose(f_actions, [
                            [i, j, [0, 0, 0]],
                            [nl1[li-1][N_LEFTMOST] - 1,
                             nl2[lj-1][N_LEFTMOST] - 1,
                             [AT_SWAP, swap_trees(nl1[i], nl1[li-1],
                                                  nl2[j], nl2[lj-1]),
                              0, nl1[i], nl1[li-1]
                              ]
                             ]
                            ]) 

                        # min between just calculed and last forest distance
                        val = f_actions[li-1][lj-1][-1][EZS_A_FCOST] + \
                              actions[i][j][-1][EZS_A_FCOST]
                        if f_actions[i][j][-1][EZS_A_FCOST] > val:
                            # concat the 2 actions list
                            f_actions[i][j] = actions[i][j][:]
                            sibl_cost = f_actions[i][j][-1][EZS_A_FCOST]
                            if li-1 > 0 and lj-1 > 0:
                                for action in f_actions[li-1][lj-1]:
                                    action = action[:]
                                    action[EZS_A_FCOST] = action[EZS_A_FCOST] +\
                                                          sibl_cost
                                    f_actions[i][j].append(action)
    
    def _mainformat(self, action_list):
        """ transform ezs output in standard format """
        # remove actions with cost = 0
        action_list = filter(lambda x: x[EZS_A_COST]!=0, action_list)
        for action in action_list:
            n_action = action #None
            # step1: transform the 3 operations SWAP, REMOVE, APPEND
            # from ezs output to SWAP, REMOVE, APPEND, UPDATE according
            # to the node and action type
            #        print '-'*80
            #        print 'action',action
##             # if the action main node have been added
##             if action[EZS_A_N1][N_TYPE] == NT_SYST:
##                 if action[EZS_A_TYPE] in (AT_APPEND, AT_SWAP):
##                     node2 = action[EZS_A_N2][N_CHILDS][0]
##                 else:
##                     try:
##                         node2 = action[EZS_A_N2]
##                     except: node2 = None
##                 n_action = [action[EZS_A_TYPE], 1, 0, action[EZS_A_N1][N_CHILDS][0], node2,
##                             node2]             
##             # action main node is from the original document
##             else: # action[EZS_A_N1][N_TYPE] != NT_SYST
            # those nodes should only be remove + append (= update)
##             if action[EZS_A_TYPE] == AT_APPEND:
##                 if action[EZS_A_N2][N_VALUE] in ['N','T','C']:
##                     delete = action[EZS_A_N2]
##                 elif action[EZS_A_N2][N_PARENT]:
##                     delete = action[EZS_A_N2][N_CHILDS][get_pos(action[EZS_A_N1][N_PARENT])]
## ##                     delete = action[EZS_A_N2][N_PARENT][N_CHILDS][get_pos(action[EZS_A_N1][N_PARENT])][N_CHILDS][0]
##                 else:
##                     # the root has changed
##                     delete = action[EZS_A_N2]
##                 # attribute node
## ##                 if action[EZS_A_N1][N_TYPE] in (NT_ATTN, NT_ATTV):
## ##                     node2 = action[EZS_A_N2][N_PARENT][N_CHILDS][0]
##                 if action[EZS_A_N1][N_TYPE] == NT_ATTN:
##                     node2 = action[EZS_A_N2][N_PARENT]
##                 elif action[EZS_A_N1][N_TYPE] == NT_ATTV:
##                     node2 = action[EZS_A_N2][N_PARENT][N_PARENT]
##                 # element node
##                 elif action[EZS_A_N1][N_TYPE] == NT_NODE:
##                     node2 = action[EZS_A_N2]#[N_CHILDS][0]
##                 # comment or textnode
##                 else: #if action[EZS_A_N1][EZS_A_TYPE] == NT_TEXT:
##                     node2 = action[EZS_A_N2] 
##                 n_action = [AT_UPDATE, 1, 0, action[EZS_A_N1], node2, delete]
        
##             # step2: transform the 4 operations SWAP, REMOVE, APPEND, UPDATE from step1
##             # output to SWAP, REMOVE, APPEND, UPDATE, INSERT_AFTER, INSERT_BEFORE and
##             # RENAME according to the nodes and action type, and convert it to list
##             if n_action:
            if n_action[EZS_A_TYPE] == AT_UPDATE:
                if n_action[EZS_A_N1][N_TYPE] in (NT_NODE, NT_ATTN) :
                    action_l = ['rename',
                                f_xpath(n_action[EZS_A_DEL]),
                                n_action[EZS_A_N1][N_VALUE]]
                else:
                    action_l = ['update',
                                f_xpath(n_action[EZS_A_DEL]),
                                n_action[EZS_A_N1][N_VALUE]]

            elif n_action[EZS_A_TYPE] == AT_SWAP:
                action_l = ['swap', n_action[EZS_A_N1], n_action[EZS_A_N2]]
            elif n_action[EZS_A_TYPE] == AT_REMOVE:
                action_l = ['remove', f_xpath(n_action[EZS_A_N1])]
            elif n_action[EZS_A_TYPE] == AT_APPEND:
                if n_action[EZS_A_N1][N_TYPE] == NT_ATTN:
                    action_l = ['append',
                                f_xpath(n_action[EZS_A_N2]),
                                n_action[EZS_A_N1]]
                elif n_action[EZS_A_N2][N_PARENT] and \
                         nb_childs(n_action[EZS_A_N2][N_PARENT]) > 1:
                    index = get_pos(n_action[EZS_A_N1][N_PARENT])
                    if index == 1 and \
                           nb_childs(n_action[EZS_A_DEL][N_PARENT]) > 1:
                        action_l = ['append-first',
                                    f_xpath(n_action[EZS_A_N1][N_PARENT][N_PARENT][N_CHILDS][0]),
                                    n_action[EZS_A_N1]]                        
                    elif index > 1:
                        action_l = ['insert-after',
                                    f_xpath(n_action[EZS_A_N1][N_PARENT][N_PARENT][N_CHILDS][index-1]), 
                                    n_action[EZS_A_N1]]                        
                    else:
                        action_l = ['append-last',
                                    f_xpath(n_action[EZS_A_DEL][N_PARENT]),
                                    n_action[EZS_A_N1]]
                else:
                    action_l = ['append', f_xpath(n_action[EZS_A_N2]), n_action[EZS_A_N1]]                
            # fully format this action
            self._formatter.add_action(action_l)

    def _post_order(self, node, nodes_list, keyroot, nodes=1):
        '''
        recursivre function which add following attributes to the tree:
        * "number",    the post ordered number of the node (integer),
        * "left most", the post ordered number of the left most child
        (or itself if none)
        * "keyroot",   a boolean (all nodes are keyroot except each
        leftmost nodes)   
        each element node is post ordered numbered
        return the current number (equal the number of nodes when all
        the tree has been processed)
        '''
        if node is not None:
            # add keyroot and leftmost attributes
            node.append(keyroot)
            node.append(nodes)
            for child in node[N_CHILDS]:
                nodes = self._post_order(child, nodes_list,
                                   child is not node[N_CHILDS][0], nodes)
            nodes_list.append(node)
            return nodes + 1
        
        return nodes



# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
provides constantes for using node and action (list) and some functions
for these objects use

    /!\  /!\ do not call index, remove or compare two node with == since a
             node is a recursive list
"""
__revision__ = '$Id: objects.py,v 1.22 2004/12/23 12:30:33 alf Exp $'

from logilab.xmldiff.misc import TRUE, FALSE
from  sys import stdout, stderr

XUPD_URI = 'http://www.xmldb.org/xupdate'
XUPD_PREFIX = 'xupdate'

################ ACTIONS #######################################################

A_DESC = 0 # string describes the action
A_N1   = 1 # node on which the action applies
A_N2   = 2 # optionnal second action argument, maybe node or value

# action types (for internal use in ezs algorithm)
AT_APPEND = 1 
AT_REMOVE = 2
AT_SWAP   = 3
AT_UPDATE = 4
AT_RENAME = 5

def actp(act):
    """ print an internal action (debugging purpose) """
    if len(act) > 2:
        if act[A_DESC][0]=='m':
            print >> stderr,  act[A_DESC], caract(act[A_N1])
            print >> stderr, '    ', caract(act[A_N2])
            print >> stderr, '    ', caract(act[-2]), act[-3], get_pos(act[-1])
        else:
            print >> stderr, act[A_DESC], caract(act[A_N1]),\
                             caract(act[A_N2]),\
                             act[A_N2][N_VALUE]
    else:
        print >> stderr, act[A_DESC], caract(act[A_N1])


################## NODES CONSTANTES ############################################

N_TYPE     = 0 # node's type
N_NAME     = 1 # node's label (to process xpath)
N_VALUE    = 2 # node's value
N_CHILDS   = 3 # nodes's childs list
N_PARENT   = 4 # node's parent
N_ISSUE    = 5 # node's total issue number
N_XNUM     = 6 # to compute node's xpath
NSIZE = 7 #number of items in a list which represent a node

# NODE TYPES
#NT_SYST = 0 # SYSTEM node (added by parser) /!\ deprecated
NT_NODE = 1 # ELEMENT node
NT_ATTN = 2 # ATTRIBUTE NAME node
NT_ATTV = 3 # ATTRIBUTE VALUE node
NT_TEXT = 4 # TEXT node
NT_COMM = 5 # COMMENT node
NT_ROOT = 6 # root node
NODES_TYPES = ('NT','NN','AN','AV','T','C', 'R') # for printing


################## OPERATIONS EDITING NODES ####################################

def link_node(parent, child):
    """ link child to his parent """
    if child:
        parent[N_CHILDS].append(child)
        child[N_PARENT] = parent

def insert_node(node, new, pos):
    """ insert child new on node at position pos (integer) """
    node[N_CHILDS].insert(pos, new)
    new[N_PARENT] = node
    i, j = 0, 1
    while i < len(node[N_CHILDS]):
        n = node[N_CHILDS][i]
        if n[N_NAME] == new[N_NAME] and n[N_TYPE] == new[N_TYPE]:
            n[N_XNUM] = j
            j += 1
        i += 1
            
def delete_node(node):
    """ delete a node from its tree """
    siblings = node[N_PARENT][N_CHILDS]
    i = get_pos(node)
    siblings.pop(i)
    node[N_PARENT] = None
    while i < len(siblings):
        n = siblings[i]
        if n[N_NAME] == node[N_NAME] and n[N_TYPE] == node[N_TYPE]:
            n[N_XNUM] -= 1
        i += 1

def rename_node(node, new_name):
    """ rename a node
    this is necessary for xpath
    """
    siblings = node[N_PARENT][N_CHILDS]
    pos = get_pos(node)
    xnum = 1
    for i in range(len(siblings)):
        n = siblings[i]
        if i < pos:
            if n[N_NAME] == new_name and n[N_TYPE] == node[N_TYPE]:
                xnum += 1
        elif i != pos:
            if n[N_NAME] == node[N_NAME] and n[N_TYPE] == node[N_TYPE]:
                n[N_XNUM] -= 1
            elif n[N_NAME] == new_name and n[N_TYPE] == node[N_TYPE]:
                n[N_XNUM] += 1
    node[N_NAME] = new_name
    node[N_XNUM] = xnum


################## OPERATIONS FORMATING NODES ##################################

def caract(node):
    """ return a string which represent the node """
    return '%s:%s (%s) %s %s' % (NODES_TYPES[node[N_TYPE]], node[N_VALUE],
                              f_xpath(node), id(node), node[N_ISSUE])
        
def f_xpath(node, x=''):
    """ compute node's xpath """
    if node[N_NAME] != '/':
        if node[N_TYPE] == NT_ATTN:
            return  f_xpath(node[N_PARENT], 
                            '/%s'%node[N_NAME][:len(node[N_NAME])-4])
        if node[N_TYPE] == NT_ATTV:
            return  f_xpath(node[N_PARENT]) #[N_PARENT], '/%s'%node[N_NAME])
        return f_xpath(node[N_PARENT], '/%s[%d]%s'%(
            node[N_NAME], node[N_XNUM], x))
    elif not x:
        return '/'
    return x

def node_repr(node):
    """ return a string which represents the given node """
    s = '%s\n' % caract(node)
    for child in node[N_CHILDS]:
        s = '%s%s' % (s, _indent(child, '  '))
    return s

def _indent(node, indent_str):
    s = '%s\-%s\n' % (indent_str, caract(node))
    if next_sibling(node) is not None:
        indent_str = '%s| ' % indent_str
    else:
        indent_str = '%s  ' % indent_str
    for child in node[N_CHILDS]:
        s = '%s%s' % (s, _indent(child, indent_str))
    return s

def xml_print(node, indent='', xupdate=0, stream=stdout):
    """
    recursive function which write the node in an xml form without the added
    nodes
    """
    if xupdate:
        _xml_print_xupdate(node, indent, stream)
    else:
        _xml_print_internal_format(node, indent, stream)

def _xml_print_internal_format(node, indent, stream):
    if node[N_TYPE] == NT_NODE:
        attrs_s = ''
        i = 0
        while i < len(node[N_CHILDS]):
            n = node[N_CHILDS][i]
            if n[N_TYPE] == NT_ATTN:
                i += 1
                attrs_s = '%s %s="%s"' % (attrs_s, n[N_VALUE],
                                          n[N_CHILDS][0][N_VALUE])
            else:
                break
        if len(node[N_CHILDS]) > i:
            stream.write('%s<%s%s>\n' % (indent, node[N_VALUE], attrs_s))
            for _curr_node in node[N_CHILDS][i:]:
                _xml_print_internal_format(_curr_node, indent + '  ', stream=stream)
            stream.write('%s</%s>\n' % (indent, node[N_VALUE]))
        else:
            stream.write('%s<%s%s/>\n' % (indent, node[N_VALUE], attrs_s))
    elif node[N_TYPE] == NT_ATTN:
        stream.write('%s<@%s>\n' % (indent, node[N_VALUE]))
        stream.write(node[N_CHILDS][0][N_VALUE] + '\n')
        stream.write('%s</%s>\n' % (indent, node[N_VALUE]))
    elif node[N_TYPE] == NT_COMM:
        stream.write('%s<!-- %s -->\n' % (indent, node[N_VALUE]))
    elif node[N_TYPE] == NT_TEXT:
        stream.write(node[N_VALUE] + '\n')
    else:
        stream.write('unknown node type',`node[N_TYPE]`)

def _xml_print_xupdate(node, indent, stream):
    # if suffix -> xupdate
    attrs_s = ' name="%s"' % node[N_VALUE]
    if node[N_TYPE] == NT_NODE:
        stream.write('%s<%s:element%s>' % (indent, XUPD_PREFIX, attrs_s))
        i = 0
        while i < len(node[N_CHILDS]):
            n = node[N_CHILDS][i]
            if n[N_TYPE] == NT_ATTN:
                stream.write('%s  <%s:attribute name="%s">' % (indent,
                                                               XUPD_PREFIX,
                                                               n[N_VALUE]))
                stream.write('%s' % n[N_CHILDS][0][N_VALUE])
                stream.write('</%s:attribute>\n' % XUPD_PREFIX)
            else:
                xml_print(n, indent = indent + '  ', stream = stream)
            i += 1
        stream.write('%s</%s:element>\n' % (indent, XUPD_PREFIX))
    elif node[N_TYPE] == NT_ATTN:
        stream.write('%s<%s:attribute%s>' % (indent, XUPD_PREFIX, attrs_s))
        stream.write(node[N_CHILDS][0][N_VALUE])
        stream.write('</%s:attribute>\n' % XUPD_PREFIX)
    elif node[N_TYPE] == NT_ATTV:
        stream.write('%s<%s:attribute name="%s">' % (indent,
                                                     XUPD_PREFIX,
                                                     node[N_PARENT][N_VALUE]))
        stream.write(node[N_VALUE])
        stream.write('</%s:attribute>\n' % XUPD_PREFIX)
    elif node[N_TYPE] == NT_COMM:
        stream.write('%s<%s:comment>' % (indent, XUPD_PREFIX))
        stream.write(node[N_VALUE])
        stream.write('</%s:comment>\n' % XUPD_PREFIX)
    elif node[N_TYPE] == NT_TEXT:
        stream.write('%s<%s:text>' % (indent, XUPD_PREFIX))
        stream.write(node[N_VALUE])
        stream.write('</%s:text>\n' % XUPD_PREFIX)

def to_dom(node, doc, uri=None, prefix=None):
    """
    recursive function to convert internal tree in an xml dom tree without
    the added nodes
    """
    if node[N_TYPE] == NT_NODE:
        dom_n = doc.createElementNS(uri, '%selement'%prefix)
        dom_n.setAttributeNS(None, 'name', node[N_VALUE])
        for n in node[N_CHILDS]: 
            if n[N_TYPE] == NT_ATTN:
                dom_n = doc.createElementNS(uri, '%sattribute'%prefix)
                v = unicode(n[N_CHILDS][0][N_VALUE], 'UTF-8')
                dom_n.setAttributeNS(None, 'name', n[N_VALUE])
                dom_n.appendChild(doc.createTextNode(v))
            else:
                dom_n.appendChild(to_dom(n, doc, uri))
    elif node[N_TYPE] == NT_ATTN:
        dom_n = doc.createElementNS(uri, '%sattribute'%prefix)
        dom_n.setAttributeNS(None, 'name', node[N_VALUE])
        v =  unicode(node[N_CHILDS][0][N_VALUE], 'UTF-8')
        dom_n.appendChild(doc.createTextNode(v))
    elif node[N_TYPE] == NT_COMM:
        dom_n = doc.createElementNS(uri, '%scomment'%prefix)
        v =  unicode(node[N_VALUE], 'UTF-8')
        dom_n.appendChild(doc.createTextNode(v))
    elif node[N_TYPE] == NT_TEXT:
        dom_n = doc.createElementNS(uri, '%stext'%prefix)
        v =  unicode(node[N_VALUE], 'UTF-8')
        dom_n.appendChild(doc.createTextNode(v))
    return dom_n


################## OPERATIONS GIVING INFOS ON NODES ############################

def get_pos(node):
    """ return the index of a node in its parent's children list

    /!\  /!\ do not call index, remove or compare two node with == since a
             node is a recursive list
    """
    try:
        childs = node[N_PARENT][N_CHILDS]
        for i in xrange(len(childs)):
            if childs[i] is node:
                return i
    except TypeError, e:
        return -1
    except ValueError, e:
        return -1

def nb_childs(node):
    """ return the number of childs (without attribute childs) of the given node
    """
    return len(filter(lambda n: n[N_CHILDS][0][N_TYPE] != NT_ATTN,
                      node[N_CHILDS]))

def nb_attrs(node):
    """ return the number of attributes of the given node """
    for i in xrange(len(node[N_CHILDS])):
        if node[N_CHILDS][i][N_TYPE] != NT_ATTN:
            break
    else:
        try:
            i += 1
        except UnboundLocalError:
            i = 0
    return i


################## MISCELLANEOUS OPERATIONS ON NODES ###########################

def next_sibling(node):
    """ return the node's right sibling """
    if node[N_PARENT] is None:
        return None
    myindex = get_pos(node)
    if len(node[N_PARENT][N_CHILDS]) > myindex+1:
        return node[N_PARENT][N_CHILDS][myindex+1]
    return None
       
def previous_sibling(node):
    """ return the node's left sibling """
    myindex = get_pos(node)
    if node[N_PARENT] and myindex > 0:
        return node[N_PARENT][N_CHILDS][myindex-1]
    return None

def get_ancestors(node, l):
    """ append to l all the ancestors from node """
    if node[N_PARENT]:
        l.append(node)
        get_ancestors(node[N_PARENT], l)
    return l

def get_labels(tree, labels, leaf_labels):
    """
    Chain all nodes with a given label l in tree T together, from left to right,
    by filling dictionnaries labels and leaf_labels (for leaf nodes).

    Label are keys pointing to a list of nodes with this type.
    Node x occurs after y in the list if x appears before y in the in-order
    traversal of T.
    /!\  /!\
    since this isn't binary tree, post order traversal (?)
    """
    if tree and tree[N_CHILDS]:
        node = tree[N_CHILDS][0]
        while node:
            get_labels(node,  labels, leaf_labels)
            node = next_sibling(node)
        labels.setdefault(NODES_TYPES[tree[N_TYPE]], []).append(tree)    
    elif tree:
        leaf_labels.setdefault(NODES_TYPES[tree[N_TYPE]], []).append(tree)

def make_bfo_list(tree):
    """ create a list with tree nodes in breadth first order """
    l, queue = [], []
    if tree:
        l.append(tree)
        if tree[N_CHILDS]:
            node = tree[N_CHILDS][0]
            while node:
                l.append(node)
                if node[N_CHILDS]:
                    queue.append(node)
                node = next_sibling(node)
                if not node and queue:
                    node = queue.pop(0)[N_CHILDS][0]
    return l

### no more used    
## def make_po_list(tree):
##     """ create a list with tree nodes in post order """
##     l, stack, poped = [], [], 0
##     if tree:
##         if tree[N_CHILDS]:
##             node = tree[N_CHILDS][0]
##             while node:
##                 if node[N_CHILDS] and not poped:
##                     stack.append(node)
##                     node = node[N_CHILDS][0]
##                 else:
##                     l.append(node)
##                     node = next_sibling(node)
##                 poped = 0
##                 if not node and stack:
##                     node = stack.pop()
##                     poped = 1
##         l.append(tree)
##     return l
## def make_preo_list(tree):
##     """ create a list with tree nodes in pre order """
##     l, stack, poped = [], [], 0
##     if tree:
##         l.append(tree)
##         if tree[N_CHILDS]:
##             node = tree[N_CHILDS][0]
##             xl.append(node)
##             while node:
##                 if node[N_CHILDS] and not poped:
##                     stack.append(node)
##                     node = node[N_CHILDS][0]
##                 else:
##                     node = next_sibling(node)
##                 l.append(node)
##                 poped = 0
##                 if not node and stack:
##                     node = stack.pop()
##                     poped = 1
##     return l
## def get_leafs(tree, l):
##     """ return a list with all leaf nodes from left to right """
##     if tree and tree[N_CHILDS]:
##         node = tree[N_CHILDS][0]
##         while node:
##             get_leafs(node, l)
##             node = next_sibling(node) 
##     elif tree:
##         l.append(tree)
## def get_issue(node, l):
##     """ append to l all the descendants from node """
##     for child in node[N_CHILDS]:
##         l.append(child)
##         if child[N_CHILDS]:
##             get_issue(child, l)
## def contains(ancestor, node):
##     """ return true if node is descendent of ancestor """
##     if node is None:
##         return FALSE
##     if ancestor is node:
##         return TRUE
##     return contains(ancestor, node[N_PARENT])
    
            
    



# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details?
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
This file contains a parser to transform xml document into an internal
tree in order to avoid adding new primitives with tree transformation

This operation represent all the document in a tree without attributes on
nodes nor text nodes, only nodes with a name and a child list

(the tree is composed by elements of type Node, defined below)
"""

__revision__ = '$Id: parser.py,v 1.22 2004/09/03 10:06:39 alf Exp $'

from logilab.xmldiff.objects import NT_ROOT, NT_NODE, NT_ATTN, NT_ATTV, \
     NT_TEXT, NT_COMM, N_TYPE, N_ISSUE, N_CHILDS, N_VALUE, link_node
from xml.sax import ContentHandler

def _inc_xpath(h, xpath):
    try:
        h[xpath] = h[xpath] + 1
    except KeyError:
        h[xpath] = 1

class SaxHandler(ContentHandler):
    """
    Sax handler to transform xml doc into basic tree
    """

    def __init__(self, normalize_space, include_comment, encoding='UTF-8'): 
        self.encoding = encoding
        self._p_stack = [[NT_ROOT, '/', '', [], None, 0, 0]]
        self._norm_sp = normalize_space or None
        self._incl_comm = include_comment or None
        self._xpath = ''
        self._h = {}
        self._n_elmt = 0

    ## method of the ContentHandler interface #################################
    def startElement(self, name, attrs):
        name = name.encode(self.encoding)
        # process xpath 
        self._xpath = "%s%s%s" % (self._xpath, '/',  name)
        _inc_xpath(self._h, self._xpath)
        # nodes construction for element
        node = [NT_NODE, name, name, [], None, self._n_elmt+1,
                self._h[self._xpath]]
        self._n_elmt += 1
        self._xpath = "%s%s%s%s" % (self._xpath, '[', self._h[self._xpath], ']')
        # nodes construction for element's attributes
        keys = attrs.keys()
        # sort attributes to avoid further moves
        keys.sort()
        for key in keys:
            key = key.encode(self.encoding)
            self._n_elmt += 2
            attr_node = [NT_ATTN, '@%sName'%key, key, [], None, 1, 0]
            link_node(node, attr_node)
            link_node(attr_node, [NT_ATTV, '@%s'%key,
                                  attrs.get(key, '').encode(self.encoding),
                                  [],None,0,0])

        link_node(self._p_stack[-1], node)
        # set current element on the top of the father stack
        self._p_stack.append(node)
        
    def endElement(self, name):
        # process xpath 
        size = len(self._xpath)
        for i in range(size):
            size = size - 1
            if self._xpath[-i - 1] == '/':
                break
        self._xpath = self._xpath[:size]
        self._p_stack[-1][N_ISSUE] = self._n_elmt - self._p_stack[-1][N_ISSUE]
        # remove last element from stack
        self._p_stack.pop()

    def characters(self, ch):
        if self._norm_sp is not None:
            ch = ' '.join(ch.split())
        if len(ch) > 0 and ch != "\n" and ch != '  ':
            ch = ch.encode(self.encoding)
            parent = self._p_stack[-1]
            # if sibling text nodes
            if parent[N_CHILDS] and parent[N_CHILDS][-1][N_TYPE] == NT_TEXT:
                n = parent[N_CHILDS][-1]
                n[N_VALUE] = n[N_VALUE] + ch
            else:
                self._n_elmt += 1
                xpath = '%s/text()' % self._xpath
                _inc_xpath(self._h, xpath)
                # nodes construction for text
                node = [NT_TEXT, 'text()', ch, [],None,0, self._h[xpath]]
                link_node(parent, node)

    ## method of the LexicalHandler interface ##################################
    def comment(self, content):
        if self._incl_comm is None:
            return
        if self._norm_sp is not None:
            content = ' '.join(content.split())
        if len(content) > 0:
            self._n_elmt += 1
            content = content.encode(self.encoding)
            xpath = '%s/comment()' % self._xpath
            _inc_xpath(self._h, xpath)
            # nodes construction for comment
            node = [NT_COMM,  'comment()', content, [], None, 0, self._h[xpath]]
            link_node(self._p_stack[-1], node)

    # methods from xml.sax.saxlib.LexicalHandler (avoid dependencie to pyxml)
    def startDTD(self, name, public_id, system_id):
        """Report the start of the DTD declarations, if the document
        has an associated DTD.

        A startEntity event will be reported before declaration events
        from the external DTD subset are reported, and this can be
        used to infer from which subset DTD declarations derive.

        name is the name of the document element type, public_id the
        public identifier of the DTD (or None if none were supplied)
        and system_id the system identfier of the external subset (or
        None if none were supplied)."""

    def endDTD(self):
        "Signals the end of DTD declarations."

    def startEntity(self, name):
        """Report the beginning of an entity.

        The start and end of the document entity is not reported. The
        start and end of the external DTD subset is reported with the
        pseudo-name '[dtd]'.

        Skipped entities will be reported through the skippedEntity
        event of the ContentHandler rather than through this event.

        name is the name of the entity. If it is a parameter entity,
        the name will begin with '%'."""

    def endEntity(self, name):
        """Reports the end of an entity. name is the name of the
        entity, and follows the same conventions as for
        startEntity."""

    def startCDATA(self):
        """Reports the beginning of a CDATA marked section.

        The contents of the CDATA marked section will be reported
        through the characters event."""

    def endCDATA(self):
        "Reports the end of a CDATA marked section."

    def get_tree(self):
        self._p_stack[0][N_ISSUE] = self._n_elmt
        return self._p_stack[0]



# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2002-2003 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr  
"""
__revision__ = '$Id: __init__.py,v 1.8 2003/09/05 16:03:57 syt Exp $'




# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
this module provides classes to format the native tree2tree output
"""
__revision__ = '$Id: format.py,v 1.28 2004/12/23 12:16:41 alf Exp $'



import types
try:
    from xml.dom import EMPTY_NAMESPACE as NO_NS
except:
    NO_NS = None
from logilab.xmldiff.objects import A_N1, A_N2, A_DESC, N_PARENT, caract, \
     xml_print, f_xpath, XUPD_PREFIX, XUPD_URI, to_dom
from sys import stdout

def get_attrs_string(attrs):
    """ extract and return a string corresponding to an attributes list """
    attr_s = []
    for attr_n, attr_v in attrs:
        attr_s.append('%s="%s" '%(attr_n, attr_v))
    return ' '.join(attr_s)

## XUPDATE FUNCTIONS ##########################################################

def open_xupdate_node(type, attrs, indent='', stream=stdout):
    """print opening xupdate tag"""
    stream.write('<%s:%s %s>' % (XUPD_PREFIX,
                                   type, get_attrs_string(attrs)))
    
def close_xupdate_node(action, indent='', stream=stdout):
    """print closing xupdate tag"""
    stream.write('</%s:%s>\n' % (XUPD_PREFIX, action))

def write_xupdate_node(type, attrs, indent='', stream=stdout):
    """print single xupdate tag"""
    stream.write('<%s:%s %s/>\n' % (XUPD_PREFIX, type,
                                      get_attrs_string(attrs)))
    

## Formatter interface ########################################################
class AbstractFormatter:
    """
    Formatter interface
    """

    def init(self, stream=stdout):
        """ method called before the begining of the tree 2 tree correction """
        self.edit_s = []
        self._stream = stream
    
    def add_action(self, action):
        """ method called when an action is added to the edit script """
        self.edit_s.append(action)

    def format_action(self, action):
        """ method called by end() to format each action in the edit script
        at least this method should be overriden
        """
        raise NotImplementedError()

    def end(self):
        """ method called at the end of the tree 2 tree correction """
        for action in self.edit_s:
            self.format_action(action)


## Internal Formatter ##########################################################

class InternalPrinter(AbstractFormatter):
    """ print actions in the internal format """
    
    def add_action(self, action):
        """
        See AbstractFormatter interface
        """
        if len(action) > 2 and type(action[A_N2]) == types.ListType:
            if type(action[A_N1]) == types.ListType:
                #swap or move node
                action[A_N1] = f_xpath(action[A_N1])
                action[A_N2] = f_xpath(action[A_N2])
        AbstractFormatter.add_action(self, action)
        
    def format_action(self, action):
        """
        See AbstractFormatter interface
        """
        if len(action) > 2 and type(action[A_N2]) == types.ListType:
            self._stream.write('[%s, %s,\n' % (action[A_DESC], action[A_N1]))
            xml_print(action[A_N2])
            self._stream.write("]\n")
        elif len(action) > 2:
            self._stream.write('[%s, %s, %s]\n' % (action[A_DESC],
                                             action[A_N1],
                                             action[A_N2]))
        else:
            self._stream.write('[%s, %s]\n' % (action[A_DESC], action[A_N1]))

## XUpdate Formatters (text / DOM) #############################################

class XUpdateMixIn:
    """ XUpdate mixin to preprocess added actions """

    def add_action(self, action):
        """
        See AbstractFormatter interface
        """
        if action[A_DESC] == 'move-first':
            # replace move-first with remove and insert (sibling nodes)
            self.edit_s.append(('remove', f_xpath(action[A_N1])))
            self.edit_s.append(('append', f_xpath(action[A_N2]), action[A_N1]))
        elif action[A_DESC] == 'move-after':
            # replace move-after with remove and insert (sibling nodes)
            self.edit_s.append(('remove', f_xpath(action[A_N1])))
            self.edit_s.append(('insert-after',
                                f_xpath(action[A_N2]),
                                action[A_N1]))
        elif action[A_DESC] == 'move-and-rename':
            # replace move-and-rename with remove and append (attribute nodes)
            self.edit_s.append(('remove', f_xpath(action[A_N1])))
            self.edit_s.append(('append',
                                f_xpath(action[A_N2][N_PARENT]),
                                action[A_N2]))
        elif action[A_DESC] == 'swap':
            # replace swap with remove and insert (sibling nodes)
            self.edit_s.append(('remove', f_xpath(action[A_N2])))
            self.edit_s.append(('insert-after',
                                f_xpath(action[A_N1]),
                                action[A_N2]))
        else:
            self.edit_s.append(action)

        
class XUpdatePrinter(XUpdateMixIn, AbstractFormatter):
    """ take the actions list in standard format and output it following
    Xupdate xml specification
    """
    def init(self, stream = stdout):
        """
        See AbstractFormatter interface
        """
        AbstractFormatter.init(self, stream)
        self._stream.write('''<?xml version="1.0"?> 
<xupdate:modifications version="1.0" xmlns:%s="%s">\n''' % (XUPD_PREFIX,
                                                            XUPD_URI))
        
    def format_action(self, action, indent='  '):
        """
        See AbstractFormatter interface
        """
        if action[A_DESC] == 'remove': 
            write_xupdate_node(action[A_DESC],
                               (('select', action[A_N1]), ),
                               indent,
                               self._stream)
        elif action[A_DESC] == 'append-last':
            open_xupdate_node('append',
                              (('select', action[A_N1]),
                               ('child', 'last()')),
                              indent,
                              self._stream)
            xml_print(action[A_N2], indent, xupdate=1, stream=self._stream)
            close_xupdate_node('append', indent, self._stream)
        elif action[A_DESC] == 'append-first':
            open_xupdate_node('append',
                              (('select', action[A_N1]),
                               ('child', 'first()')),
                              indent,
                              self._stream)
            xml_print(action[A_N2], indent, xupdate=1, stream=self._stream)
            close_xupdate_node('append', indent, self._stream)
        elif action[A_DESC] in ['append', 'insert-after']:
            open_xupdate_node(action[A_DESC],
                              (('select', action[A_N1]), ),
                              indent, self._stream)
            xml_print(action[A_N2], indent, xupdate=1, stream=self._stream)
            close_xupdate_node(action[A_DESC], indent, self._stream)
        elif action[A_DESC] == 'rename': 
            open_xupdate_node(action[A_DESC],
                              (('select', action[A_N1]), ),
                              indent, self._stream)
            self._stream.write(action[A_N2])
            close_xupdate_node(action[A_DESC], indent, self._stream)
        else:
            open_xupdate_node(action[A_DESC],
                              (('select', action[A_N1]), ),
                              indent, self._stream)
            self._stream.write(action[A_N2])
            close_xupdate_node(action[A_DESC], indent, self._stream)
        self._stream.write('\n')
        self._stream.flush()

    def end(self):
        """
        See AbstractFormatter interface
        """
        AbstractFormatter.end(self)
        self._stream.write('</%s:modifications>'%XUPD_PREFIX)


class DOMXUpdateFormatter(XUpdateMixIn, AbstractFormatter):
    """ take the actions list in standard format and return a dom tree
    which follow Xupdate xml specification (without xupdate namespace)
    dom tree is append to doc (DOM Document node)
    """
    def __init__(self, doc, encoding='UTF-8'):
        """
        Instance attributes are doc and encoding
        """
        self.doc = doc
        self.encoding = encoding
        
    def init(self):
        """
        See AbstractFormatter interface
        """
        AbstractFormatter.init(self)
        output = self.doc.createElementNS(XUPD_URI,
                                          '%s:modifications'%XUPD_PREFIX)
        output.setAttributeNS(NO_NS, 'version', '1.0')
        self.output = output
        
    def format_action(self, action):
        """
        See AbstractFormatter interface
        """
        doc = self.doc
        if action[A_DESC] == 'remove':
            node = doc.createElementNS(XUPD_URI,
                                       '%s:%s' % (XUPD_PREFIX, action[A_DESC]))
            node.setAttributeNS(NO_NS, 'select', action[A_N1])
        elif action[A_DESC] == 'append-first':
            node = doc.createElementNS(XUPD_URI,
                                       '%s:%s'% (XUPD_PREFIX, 'append'))
            node.setAttributeNS(NO_NS, 'select', action[A_N1])
            node.setAttributeNS(NO_NS, 'child', 'first()')
            node.appendChild(to_dom(action[A_N2], doc, XUPD_URI, XUPD_PREFIX))
        elif action[A_DESC] == 'append-last':
            node = doc.createElementNS(XUPD_URI,
                                       '%s:%s' % (XUPD_PREFIX, 'append'))
            node.setAttributeNS(NO_NS, 'select', action[A_N1])
            node.setAttributeNS(NO_NS, 'child', 'last()')
            node.appendChild(to_dom(action[A_N2], doc, XUPD_URI, XUPD_PREFIX))
        elif action[A_DESC] in ['append', 'insert-after', 'insert-before']:
            node = doc.createElementNS(XUPD_URI,
                                       '%s:%s' % (XUPD_PREFIX, action[A_DESC]))
            node.setAttributeNS(NO_NS, 'select', action[A_N1])
            node.appendChild(to_dom(action[A_N2], doc, XUPD_URI, XUPD_PREFIX))
        elif action[A_DESC] == 'rename': 
            node = doc.createElementNS(XUPD_URI,
                                       '%s:%s' %(XUPD_PREFIX, action[A_DESC]))
            node.setAttributeNS(NO_NS, 'name', action[A_N1])
            v = unicode(action[A_N2], self.encoding)
            node.appendChild(doc.createTextNode(v))
        else:
            node = doc.createElementNS(XUPD_URI,
                                       '%s:%s' % (XUPD_PREFIX, action[A_DESC]))
            node.setAttributeNS(NO_NS, 'select', action[A_N1])
            v = unicode(action[A_N2], self.encoding)
            node.appendChild(doc.createTextNode(v))
        # append xupdate node
        self.output.appendChild(node)




""" Provides functions for converting DOM tree or xml file in order to process
it with xmldiff functions. """ 
# Copyright (c) 2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
__revision__ = '$Id: input.py,v 1.14 2003/09/27 14:46:29 nico Exp $'

def tree_from_stream(stream, 
                     norm_sp=1, ext_ges=0, ext_pes=0, include_comment=1,
                     encoding='UTF-8', html=0):
    """
    create internal tree from xml stream (open file or IOString)
    if norm_sp = 1, normalize space and new line
    """
    from xml.sax import make_parser, SAXNotRecognizedException
    from xml.sax.handler import feature_namespaces, feature_external_ges, \
         feature_external_pes, property_lexical_handler
    from logilab.xmldiff.parser import SaxHandler
    handler = SaxHandler(norm_sp, include_comment, encoding)
    if html:
        parser = make_parser(["xml.sax.drivers2.drv_sgmlop_html"])
    else:
        parser = make_parser()
        # do not perform Namespace processing 
        parser.setFeature(feature_namespaces, 0)
    # do not include any external entities
    try:
        parser.setFeature(feature_external_ges, ext_ges)
        #xml.sax._exceptions.
    except SAXNotRecognizedException:
        print 'Unable to set feature external ges'
    try:
        parser.setFeature(feature_external_pes, ext_pes)
        #xml.sax._exceptions.
    except SAXNotRecognizedException:
        print 'Unable to set feature external pes'
    # add lexical handler for comments,  entities, dtd and cdata
    parser.setProperty(property_lexical_handler, handler)
    parser.setContentHandler(handler)
    parser.parse(stream)
    return handler.get_tree()


def tree_from_dom(root, ezs=0):
    """ create internal tree from DOM subtree """
    from xml.dom.ext.Dom2Sax import Dom2SaxParser 
    from xml.sax.handler import feature_namespaces, property_lexical_handler
    #from parser import DomParser
    parser = Dom2SaxParser()
    from logilab.xmldiff.parser import SaxHandler
    handler = SaxHandler(normalize_space=0, include_comment=1)
    # do not perform Namespace processing 
    parser.setFeature(feature_namespaces, 0)
    # add lexical handler for comments,  entities, dtd and cdata
    parser.setProperty(property_lexical_handler, handler)
    parser.setContentHandler(handler)
    parser.parse(root)
    return handler.get_tree()


if __name__ == '__main__':
    from xml.dom.ext import StripXml, PrettyPrint
    from xml.dom.ext.reader.Sax2 import Reader
    import sys
    reader = Reader()
    file = open(sys.argv[1],'r')
    fragment = reader.fromStream(file)
    d = StripXml(fragment)
    file.close()
    tree = tree_from_dom(d)
    file = open(sys.argv[2],'r')
    fragment = reader.fromStream(file)
    d = StripXml(fragment)
    file.close()
    tree2 = tree_from_dom(d)
    from logilab.xmldiff.objects import repr
    print 'Source tree', repr(tree)
    print 'Destination tree', repr(tree2)
    #from ezs import EzsCorrector
    #strategy = EzsCorrector()
    from logilab.xmldiff.fmes import FmesCorrector
    strategy = FmesCorrector(0.59, 0.5)
    #from ezs import process
    actions = strategy.process_trees(tree, tree2)
    from logilab.xmldiff.format import xupdate_dom
    PrettyPrint(
        xupdate_dom(
        reader.fromString('<root/>'),
        actions))



"""
miscellaneous functions
"""
# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

__revision__ = '$Id: misc.py,v 1.10 2003/10/02 10:07:49 syt Exp $'

##### interface ###        
TRUE  = 1
FALSE = 0

def process_dirs(dir1, dir2, recursive):
    """
    function which return common, added, deleted file from dir1 to dir2
    and remove initial directory from paths
    """
    dir1 = normalize_dir(dir1)
    dir2 = normalize_dir(dir2)
    common, deleted, added = _process_dirs(dir1, dir2, recursive)
    # remove prefix
    deleted[0] = map(_remove_prefix(len(dir1)),  deleted[0])
    deleted[1] = map(_remove_prefix(len(dir1)),  deleted[1])
    added[0] = map(_remove_prefix(len(dir2)),  added[0])
    added[1] = map(_remove_prefix(len(dir2)),  added[1])
    common[0] = map(_remove_prefix(len(dir1)), common[0])
    common[1] = map(_remove_prefix(len(dir1)), common[1])
    return common, deleted, added

def divide_files(dir):
    """ return a list with subdir of dir and another one with files """
    import os
    dirs=[]
    regs=[]
    #os.path.normpath(dir)
    for filename in os.listdir(dir):
        if os.path.isfile(os.path.join(dir, filename)):
            regs.append(filename)
        elif os.path.isdir(os.path.join(dir, filename)):
            dirs.append(filename)
    return dirs, regs

def extract(list1, list2):
    """ extract common, added, deleted item from list1 to list2 """
    common, deleted, added = [],[],[]
    for item in list1:
        if item in list2:
            common.append(item)
            list2.remove(item)
        else:
            deleted.append(item)
    for item in list2:
        if not item in list1:
            added.append(item)
    return common, deleted, added

def union(list1, list2):
    """ return list1 union list2 """
    tmp = {}
    for i in list1:
        tmp[i]=1
    for i in list2:
        tmp[i]=1
    return tmp.keys()

def intersection(list1, list2):
    """ return common items in list1 and list2 """
    tmp, l = {}, []
    for i in list1:
        tmp[i]=1
    for i in list2:
        if tmp.has_key(i):
            l.append(i)
    return l

def init_matrix(nbrows, nbcols, default_value):
    """
    return a 2d matrix list2d[nbrows][nbcols] initialised with
    default_value (carefull for side effect with references...)
    """
    list2d = []    
    for i in range(nbrows):
        tmpl = []
        for j in range(nbcols):
            tmpl.append(default_value)
        list2d.append(tmpl)
    return list2d

def in_ref(list, item):
    """ return true if list contains a reference on item """
    for it in list:
        if it is item:
            return TRUE
    return FALSE

def index_ref(list, item):
    """
    return the index of item in list by reference comparison
    raise Exception if not found
    """
    index = 0
    for it in list:
        if it is item:
            return index
        index += 1
    raise Exception('No item '+item)

def list_print (list, s1='', s2=''):
    for item in list:
        if item:
            print '-'*80
            print s1, item, s2

def append_list(list, a_list):
    """ append a_list items to list """ 
    for item in a_list:
        list.append(item)

##### private ###
from os import sep

def normalize_dir(directory):
    """remove trailing path separator from the directory name
    """
    while directory[-1] == sep:
        directory = directory[:-1]
    return directory

def _process_dirs(dir1, dir2, recursive):
    """
    function which return common, added, deleted file from dir1 to dir2
    if recursive, enter in subdirectory
    """
    #/!\ sym links /!\#
    
    # extract common files and directory
    d_list1, f_list1 = divide_files(dir1)
    d_list2, f_list2 = divide_files(dir2)
    common, deleted, added = [[],[]], [[],[]], [[],[]]
    common[0], deleted[0], added[0] = extract(f_list1, f_list2)
    common[1], deleted[1], added[1] = extract(d_list1, d_list2)
    # add prefix 
    deleted[0] = map(_add_prefix(dir1),  deleted[0])
    deleted[1] = map(_add_prefix(dir1),  deleted[1])
    added[0] = map(_add_prefix(dir2),  added[0])
    added[1] = map(_add_prefix(dir2),  added[1])
    common[0] = map(_add_prefix(dir1), common[0])
    if recursive:
        import os
        # for all common subdirs
        for dir in common[1]:
            if dir == 'CVS':
                continue
            # recursion
            comm, dele, adde = _process_dirs(os.path.join(dir1, dir),
                                             os.path.join(dir2, dir),
                                             recursive)
            # add subdir items
            append_list(deleted[0], dele[0]); append_list(deleted[1], dele[1])
            append_list(added[0], adde[0]); append_list(added[1], adde[1])
            append_list(common[0], comm[0]); 
    return common, deleted, added

def _remove_prefix(prfx_size):
    """ return a function to add remove with map() """
    return lambda s, len=prfx_size: s[len+1:]

def _add_prefix(prefix):
    """ return a function to add prefix with map() """
    import os
    return lambda s, prfx=prefix,join=os.path.join: join(prfx, s)




# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
"""
 this file provides the fast match / edit script (fmes) tree to tree correction
 algorithm as described in "Change detection in hierarchically structured
 information" by S. Chawathe, A. Rajaraman, H. Garcia-Molina and J. Widom
 ([CRGMW95])
"""

__revision__ = "$Id: fmes.py,v 1.20 2004/09/03 10:06:38 alf Exp $"

from logilab.xmldiff.objects import NT_ROOT, NT_NODE, NT_ATTN, NT_ATTV, \
     NT_TEXT, NT_COMM, N_TYPE, N_NAME, N_VALUE, N_CHILDS, N_PARENT, N_ISSUE, \
     N_XNUM, NSIZE, A_DESC, A_N1, A_N2, FALSE, TRUE, \
     node_repr, get_labels, get_ancestors, caract, make_bfo_list, \
     insert_node, delete_node, rename_node, get_pos, \
     f_xpath, nb_attrs, xml_print
from logilab.xmldiff.mydifflib import lcs2, quick_ratio
from logilab.xmldiff.misc import intersection, in_ref, index_ref
# c extensions
from logilab.xmldiff.maplookup import has_couple , partner, fmes_init, \
     fmes_node_equal, match_end, fmes_end

# node's attributes for fmes algorithm
N_INORDER = NSIZE
N_MAPPED = N_INORDER + 1

def _init_tree(tree, map_attr=None):
    """ recursively append N_INORDER attribute to tree
    optionnaly add the N_MAPPED attribute (for node from tree 1)
    """
    tree.append(FALSE)
    if not map_attr is None:
        tree.append(FALSE)
    for child in tree[N_CHILDS]:
        _init_tree(child, map_attr)

## FMES TREE 2 TREE ALGORITHM #################################################
class FmesCorrector:
    """
    Fast Match / Edit Script implementation

    See [CRGMW95] for reference.
    """
    
    def __init__(self, formatter, f=0.6, t=0.5): # f=0,59
        # algorithm parameters
        if f>1 or f<0 or t>1 or t<0.5:
            raise Exception('Invalid parameters:  1 > f > 0 and 1 > t > 0.5')
        self.F = f
        self.T = t
        self._formatter = formatter

    def process_trees(self, tree1, tree2):
        """
        Process the two trees
        """
        # add needed attribute (INORDER)
        _init_tree(tree1, map_attr=1)
        _init_tree(tree2)
##         print '**** TREE 2'
##         print node_repr(tree2)
##         print '**** TREE 1'
##         print node_repr(tree1)
        # attributes initialisation
        self._mapping = []  # empty mapping
        self.add_action = self._formatter.add_action
        self._d1, self._d2 = {}, {}
        # give references to the C extensions specific to fmes
        fmes_init(self._mapping, self._d1, self._d2, self.T)
        self._dict = {}
        self._tmp_attrs_dict = {}
        self._pending = []
        self._formatter.init()
        # step 0: mapping
        self._fast_match(tree1, tree2)
        # free matching variables
        match_end()
        del self._d1
        del self._d2
        # step 1: breadth first search tree2
        self._fmes_step1(tree2, tree1)
        # step 2: post order traversal tree1
        self._fmes_step2(tree1, tree2)
        # step 3: rename tmp attributes
        for tmp_name, real_name in self._tmp_attrs_dict.items():
            self.add_action(['rename','//%s'%tmp_name, real_name])
        # free mapping ref in C extensions
        fmes_end()
        self._formatter.end()

    ## Private functions ######################################################
    def _fast_match(self, tree1, tree2):
        """ the fast match algorithm
        try to resolve the 'good matching problem' """
        labl1, labl2 = {}, {}
        leaf_labl1, leaf_labl2 = {}, {}
        # chain all nodes with a given label l in tree T together
        get_labels(tree1, labl1, leaf_labl1)
        get_labels(tree2, labl2, leaf_labl2)
        # do the matching job
        self._match(leaf_labl1, leaf_labl2, self._l_equal)
        # remove roots ('/') from labels
        del labl1['R']
        del labl2['R']
        # append roots to mapping
        self._mapping.append((tree1, tree2))
        # mark node as mapped
        tree1[N_MAPPED] = TRUE
        self._match(labl1, labl2, fmes_node_equal)#self._n_equal
    
    def _match(self, lab_l1, lab_l2, equal):
        """do the actual matching"""
        d1, d2 = self._d1, self._d2
        mapping = self._mapping
        # for each leaf label in both tree1 and tree2
        l = intersection(lab_l1.keys(), lab_l2.keys())
        # sort list to avoid differences between python version
        l.sort()
        for label in l:
            s1 = lab_l1[label]
            s2 = lab_l2[label]
            # compute the longest common subsequence
            common = lcs2(s1, s2, equal)
            # for each pair of nodes (x,y) in the lcs
            for x, y in common:
                # add (x,y) to the mapping
                mapping.append((x, y))
                # mark node from tree 1 as mapped
                x[N_MAPPED] = TRUE
                # fill the mapping cache
                for n in get_ancestors(x, []):
                    d1[(id(n), id(x))] = 1
                for n in get_ancestors(y, []):
                    d2[(id(n), id(y))] = 1

    def _fmes_step1(self, tree2, tree1):
        """ first step of the edit script algorithm
        combines the update, insert, align and move phases
        """
        mapping = self._mapping
        fp = self._find_pos
        al = self._align_children
        _partner = partner
        # x the current node in the breadth-first order traversal
        for x in make_bfo_list(tree2):
            y = x[N_PARENT]
            z = _partner(1, y)
            w = _partner(1, x)
            # insert
            if not w:
                todo = 1
                # avoid to add existing attribute node
                if x[N_TYPE] == NT_ATTN:
                    for w in z[N_CHILDS]:
                        if w[N_TYPE] != NT_ATTN:
                            break
                        elif w[N_VALUE] == x[N_VALUE]:
                            ## FIXME: what if w or w[N_CHILDS][0] yet mapped ??
                            if not w[N_MAPPED]:
##                                 old_value = x[N_VALUE]
##                                 x[N_VALUE] = 'xmldiff-%s'%old_value
##                                 self._tmp_attrs_dict[x[N_VALUE]] = old_value
##                                 old_x = _partner(0, w)
##                                 i = 0
##                                 for i in range(len(mapping)):
##                                     if mapping[i][0] is w:
##                                         print mapping[i][1]
##                                         mapping[i][1][N_MAPPED] = FALSE
##                                         mapping.pop(i)
##                                         break
##                             else: 
                                todo = None
                                w[N_MAPPED] = TRUE
                                mapping.append((w, x))
                                # print 'delete 1'
                                # if not w[N_CHILDS][0]:
                                delete_node(w[N_CHILDS][0])
                            break
                        
                if todo is not None:
                    x[N_INORDER] = TRUE
                    k = fp(x)
                    # w = copy(x)
                    w = x[:]
                    w[N_CHILDS] = []
                    w.append(TRUE) # <-> w[N_MAPPED] = TRUE
                    mapping.append((w, x))
                    # avoid coalescing two text nodes
                    if w[N_TYPE] == NT_TEXT:
                        k = self._before_insert_text(z, w, k)
                    # real insert on tree 1
                    insert_node(z, w, k)
                    # make actions on subtree
                    self._dict[id(w)] = ww = w[:]
                    ww[N_CHILDS] = []
                    # preformat action
                    if not self._dict.has_key(id(z)):
                        if w[N_TYPE] == NT_ATTV:
                            action = ['update', f_xpath(z), w[N_VALUE]]
                        elif w[N_TYPE] == NT_ATTN:
                            action = ['append', f_xpath(z), ww]
                        elif z[N_TYPE] == NT_ROOT:
                            action = ['append-first', '/', ww]
                        else:
                            k = get_pos(w)
                            if k <= nb_attrs(z):
                                action = ['append-first',
                                          f_xpath(z), ww]
                            else:
                                action = ['insert-after',
                                          f_xpath(z[N_CHILDS][k-1]), ww]
                        self.add_action(action)
                    else:
                        insert_node(self._dict[id(z)], ww, k)
            elif x[N_NAME] != '/':
                v = w[N_PARENT]
                # update
                if w[N_VALUE] != x[N_VALUE]:
                    # format action
                    if w[N_TYPE] == NT_NODE:
                        self.add_action(['rename', f_xpath(w), x[N_VALUE]])
                    elif w[N_TYPE] == NT_ATTN:
                        attr_name = self._before_attribute(w[N_PARENT], w,
                                                           x[N_VALUE])
                        self.add_action(['rename', f_xpath(w), attr_name])
                        x[N_NAME] = '@%sName' % attr_name
                        x[N_VALUE] = attr_name
                    else:
                        self.add_action(['update', f_xpath(w), x[N_VALUE]])
                    # real update on t1
                    w[N_VALUE] = x[N_VALUE]
                    # this is necessary for xpath
                    rename_node(w, x[N_NAME])
                # move x if parents not mapped together
                if not has_couple(v, y):
                    x[N_INORDER] = TRUE
                    k = fp(x)
                    self._make_move(w, z, k)
            # align children
            al(w, x)
#            print 'after', node_repr(tree1)

        
    def _fmes_step2(self, tree1, tree2):
        """ the delete_node phase of the edit script algorithm

        instead of the standard algorithm, walk on tree1 in pre order and
        add a remove action on node not marked as mapped.
        Avoiding recursion on these node allow to extract remove on subtree
        instead of leaf

        do not use next_sibling for performance issue
        """
        stack = []
        i = 0
        node = tree1
        while node is not None:
            if node[N_MAPPED] != TRUE:
                if node[N_PARENT] and len(node[N_PARENT][N_CHILDS]) > i+1:
                    next_node = node[N_PARENT][N_CHILDS][i+1]
                    # if next node is a text node to remove, switch actions
                    if next_node[N_TYPE] == NT_TEXT and \
                       next_node[N_MAPPED] != TRUE:
                        self.add_action(['remove', f_xpath(next_node)])
                        delete_node(next_node)
                        try:
                            next_node = node[N_PARENT][N_CHILDS][i+1]
                        except:
                            next_node = None
                else:
                    next_node = None
                self.add_action(['remove', f_xpath(node)])
                delete_node(node)
                node = next_node
            elif node[N_CHILDS]:
                # push next sibbling on the stack
                if node[N_PARENT] and len(node[N_PARENT][N_CHILDS]) > i+1 :
                    stack.append((node[N_PARENT][N_CHILDS][i+1], i+1))
                node = node[N_CHILDS][0]
                i = 0
            elif node[N_PARENT] and len(node[N_PARENT][N_CHILDS]) > i+1:
                i += 1
                node = node[N_PARENT][N_CHILDS][i] #next_sibling(node)
            else:
                node = None
            if node is None and stack:
                node, i = stack.pop()

        
    def _align_children(self, w, x):
        """ align children to correct misaligned nodes
        """
        _partner = partner
        # mark all children of w an d as "out of order"
        self._childs_out_of_order(w)
        self._childs_out_of_order(x)
        # s1: children of w whose partner is children of x
        s1 = [n for n in w[N_CHILDS] if in_ref(x[N_CHILDS], _partner(0,n))]
        # s2: children of x whose partners are children of w
        s2 = [n for n in x[N_CHILDS] if in_ref(w[N_CHILDS], _partner(1,n))]
        # compute the longest common subsequence
        s = lcs2(s1, s2, has_couple)
        # mark each (a,b) from lcs in order
        for a, b in s:
            a[N_INORDER] = b[N_INORDER] = TRUE
            s1.pop(index_ref(s1, a))
        # s: a E T1, b E T2, (a,b) E M, (a;b) not E s
        for a in s1:
            b = _partner(0, a)
            # mark a and b in order
            a[N_INORDER] = b[N_INORDER] = TRUE
            k = self._find_pos(b)
            self._make_move(a, w, k)

    def _find_pos(self, x):
        """ find the position of a node in the destination tree (tree2)

        do not use previous_sibling for performance issue
        """
        y = x[N_PARENT]
        # if x is the leftmost child of y in order, return 1
        for v in y[N_CHILDS]:
            if v[N_INORDER]:
                if v is x:
                    # return 0 instead of 1 here since the first element of a
                    # list have index 0
                    return 0
                break
        # looking for rightmost left sibling of y INORDER
        i = get_pos(x) - 1
        while i >= 0:
            v = y[N_CHILDS][i]
            if v[N_INORDER]:
                break
            i -= 1
        u = partner(1, v)
        if not u is None:
            return get_pos(u)+1

    def _make_move(self, n1, n2, k):
        # avoid coalescing two text nodes
        act_node = self._before_delete_node(n1)
        if act_node is not None and act_node[0] is n2 and act_node[1] < k:
            k += 1
        if n1[N_TYPE] == NT_TEXT:
            k = self._before_insert_text(n2, n1, k)
        elif n1[N_TYPE] == NT_ATTN: 
            # avoid to move an attribute node from a place to another on
            # the same node
            if not n1[N_PARENT] is n2:
                n1_xpath = f_xpath(n1)
                old_name = n1[N_VALUE]
                new_name = self._before_attribute(n2, n1)
                if new_name != old_name:
                    self.add_action(['remove', f_xpath(n1)])
                    n1[N_NAME] = '@%sName' % new_name
                    n1[N_VALUE] = new_name
                    self.add_action(['append', f_xpath(n2), n1])
                else:
                    self.add_action(['move', n1, n2])
                    
        elif k <= nb_attrs(n2):
            self.add_action(['move-first', n1, n2])
        else:
            self.add_action(['move-after', n1, n2[N_CHILDS][k-1]])
        # real move
        delete_node(n1)
        insert_node(n2, n1, k)

    def _before_attribute(self, parent_node, attr_node, new_name=None):
        attr_name = new_name or attr_node[N_VALUE]
        for w in parent_node[N_CHILDS]:
            if w[N_TYPE] != NT_ATTN:
                break
            if w[N_VALUE] == attr_name:
                new_name = 'LogilabXmldiffTmpAttr%s'%attr_name.replace(':',
                                                                       '_')
                self._tmp_attrs_dict[new_name] = attr_name
                return new_name
        return attr_name


    FAKE_TAG = [NT_NODE, 'LogilabXMLDIFFFAKETag', 'LogilabXMLDIFFFAKETag',
                [], None, 0, 0, TRUE, FALSE]

    def _before_insert_text(self, parent, new_text, k):
        """ check if a text node that will be remove has two sibbling text
        nodes to avoid coalescing two text nodes
        """
        if k > 1:
            if parent[N_CHILDS][k-1][N_TYPE] == NT_TEXT:
                tag = self.FAKE_TAG[:]
                self.add_action(['insert-after',
                                 f_xpath(parent[N_CHILDS][k-1]), tag])
                insert_node(parent, tag, k)
                return k+1
        if k < len(parent[N_CHILDS]):
            if parent[N_CHILDS][k][N_TYPE] == NT_TEXT:
                tag = self.FAKE_TAG[:]
                if k <= nb_attrs(parent):
                    self.add_action(['append-first', f_xpath(parent), tag])
                else:
                    self.add_action(['insert-after',
                                     f_xpath(parent[N_CHILDS][k]), tag])
                insert_node(parent, tag, k)                
        return k
    
    def _before_delete_node(self, node):
        """ check if a text node will be inserted with a sibbling text node to
        avoid coalescing two text nodes
        """
        k = get_pos(node)
        parent = node[N_PARENT]
        if k >= 1 and k+1 < len(parent[N_CHILDS]):
            if parent[N_CHILDS][k-1][N_TYPE] == NT_TEXT and \
               parent[N_CHILDS][k+1][N_TYPE] == NT_TEXT:
                tag = self.FAKE_TAG[:]
                self.add_action(['insert-after',
                                 f_xpath(parent[N_CHILDS][k-1]), tag])
                insert_node(parent, tag, k)
                return parent, k
        return None
    
    def _childs_out_of_order(self, subtree):
        """ initialisation function : tag all the subtree as unordered """
        for child in subtree[N_CHILDS]:
            child[N_INORDER] = FALSE
            self._childs_out_of_order(child)

    def _l_equal(self, n1, n2):
        """ function to compare leafs during mapping """
        ratio = quick_ratio(n1[N_VALUE], n2[N_VALUE])
        if ratio > self.F:
#            print 'MATCH (%s): %s / %s' %(ratio, n1[N_VALUE],n2[N_VALUE])
            return TRUE
#        print 'UNMATCH (%s): %s / %s' %(ratio, n1[N_VALUE],n2[N_VALUE])
        return FALSE
    


try:
    import os
    if os.environ.get('PYLINT_IMPORT') != '1': # avoid erros with pylint
        import psyco
        psyco.bind(FmesCorrector._fmes_step1)
        psyco.bind(FmesCorrector._align_children)
##         psyco.bind(FmesCorrector._fmes_step2)
        psyco.bind(FmesCorrector._match)
        psyco.bind(FmesCorrector._find_pos)
except Exception, e:
    pass



from __future__ import nested_scopes
"""
xmldiff non regression test
"""
__revision__ = "$Id: regrtest.py,v 1.17 2004/12/22 16:10:15 alf Exp $"

from os.path import join, basename
from cStringIO import StringIO
import sys
import os
import unittest
import glob

from logilab.xmldiff import xmldiff

DATA_DIR = 'data'

class BaseTest(unittest.TestCase):
    def check_output(self, options, expected):
        try:
            output = os.popen('%s %s %s' % (sys.executable,
                                            xmldiff.__file__,
                                            ' '.join(options)))
        except SystemExit:
            pass
        data = output.read().strip()
        output.close()
        self.assertEqual(data, expected, '%s:\n%r != %r' %
                         (self.name, data, expected) )
        

class DiffTest(BaseTest):
    
    def test_known(self):
        old = self.data['old']
        new = self.data['new']
        for options, res_file in self.data['result']:
            options = options + [old, new]
            f = open(res_file)
            expected = f.read().strip()
            f.close()
            self.check_output(options, expected)


class RecursiveDiffTest(BaseTest):
    name = 'RecursiveDiffTest'
    def test(self):
        options = ['-r', join(DATA_DIR, 'dir1'), join(DATA_DIR, 'dir2')]
        expected = """--------------------------------------------------------------------------------
FILE: onlyindir1.xml deleted
--------------------------------------------------------------------------------
FILE: onlyindir2.xml added
--------------------------------------------------------------------------------
FILE: inbothdir.xml"""
        self.check_output(options, expected)

def make_tests():
    """generate tests classes from test info
    
    return the list of generated test classes
    """
    tests_files = glob.glob(join(DATA_DIR, '*.xml')) + glob.glob(join(DATA_DIR, '*_result')) + glob.glob(join(DATA_DIR, '*_result_xupdate'))
    tests = {}
    # regroup test files
    for filename in tests_files:
        base = basename(filename)
        name = base[:6]
        filetype = base[7:]
        if filetype == '1.xml':
            tests.setdefault(name, {})['old'] = filename
        elif filetype == '2.xml':
            tests.setdefault(name, {})['new'] = filename
        else:
            options = filetype.split('_')[:-1]
            tests.setdefault(name, {}).setdefault('result', []).append(
                [options, filename])
                    
    result = []
    for t_name, t_dict in tests.items():
        try:
            old = t_dict['old']
            new = t_dict['new']
            res_data = t_dict['result']
        except KeyError, e:
            msg = '** missing files in %s (%s)' % (t_name, e)
            print >>sys.stderr, msg
            continue
            
        class DiffTestSubclass(DiffTest):
            name = t_name
            data = t_dict
                
        result.append(DiffTestSubclass)
    return result


    
def suite():
    return unittest.TestSuite([unittest.makeSuite(test)
                               for test in make_tests() + [RecursiveDiffTest]])

def Run(runner=None):
    testsuite = suite()
    if runner is None:
        runner = unittest.TextTestRunner()
    return runner.run(testsuite)

   
if __name__ == '__main__':
    Run()



#!/usr/bin/python
# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

__revision__ = '$Id: xmldiff.py,v 1.35 2004/12/23 12:29:43 alf Exp $'

import sys

def usage(pgm):
    """
    Print usage
    """
    print 'USAGE:'
    print "\t"+pgm, '[OPTIONS] from_file to_file'
    print "\t"+pgm, '[OPTIONS] [-r] from_directory to_directory'
    print """OPTIONS: 
  -h, --help
     display this help message and exit.
  -V, --version
     display version number and exit

  -H, --html
     input files are HTML instead of XML
  -r, --recursive
     when comparing directories, recursively compare any
     subdirectories found.
     
  -x, --xupdate
     display output following the Xupdate xml specification 
     (see http://www.xmldb.org/xupdate/xupdate-wd.html#N19b1de).
  -e encoding, --encoding=encoding
     specify the encoding to use for output. Default is UTF-8

  -n, --not-normalize-spaces
     do not normalize spaces and new lines in text and comment nodes.
  -c, --exclude-comments
     do not process comment nodes
  -g, --ext-ges
     include all external general (text) entities. 
  -p, --ext-pes
     include all external parameter entities, including the external DTD
     subset. 

  --profile=file
     display an execution profile (run slower with this option),
     profile saved to file (binarie form).
"""
##   -z, --ezs
##      use the extended Zhang and Shasha algorithm, much slower
##      but with the best results (only for small documents) 

def process_files(file1, file2, norm_sp, xupd, ezs, verbose,
                  ext_ges, ext_pes, include_comment, encoding,
                  html):
    """
    Computes the diff between two files.
    """
    from xml.sax import SAXParseException
    try:
        fh1, fh2 = open(file1, 'r'), open(file2, 'r')
    except IOError, msg :
        sys.stderr.write(str(msg) + '\n')
        return -1
    # convert xml files to tree
    try:
        from logilab.xmldiff.input import tree_from_stream
        tree1 = tree_from_stream(fh1, norm_sp, ext_ges,
                                 ext_pes, include_comment,
                                 encoding, html)
        tree2 = tree_from_stream(fh2, norm_sp, ext_ges,
                                 ext_pes, include_comment,
                                 encoding, html)
        fh1.close ()
        fh2.close ()
    except SAXParseException, msg:
        print msg
        return -1

    if verbose:
        from logilab.xmldiff.objects import repr, N_ISSUE, N_CHILDS
        print "Source tree\n", repr(tree1)
        print "Destination tree\n", repr(tree2)
        print 'Source tree has', tree1[N_ISSUE], 'nodes'
        print 'Destination tree has', tree2[N_ISSUE], 'nodes'
    # output formatter
    if xupd:
        from logilab.xmldiff.format import XUpdatePrinter
        formatter = XUpdatePrinter()
    else:
        from logilab.xmldiff.format import InternalPrinter
        formatter = InternalPrinter()
    # choose and apply tree to tree algorithm
    if ezs:
        from logilab.xmldiff.ezs import EzsCorrector
        strategy = EzsCorrector(formatter)
    else:
        from logilab.xmldiff.fmes import FmesCorrector
        #import gc
        #gc.set_debug(gc.DEBUG_LEAK|gc.DEBUG_STATS)
        strategy = FmesCorrector(formatter)
    strategy.process_trees(tree1, tree2)
    return len(formatter.edit_s)

def run(*args):
    """
    Main. To be called with list of command-line arguments.

    FIXME: use optparse and remove usage() ?
    """
    import os, getopt
    pgm = args[0]
    s_opt = 'Hrncgpe:xzhvV'
    l_opt = ['html', 'recursive',
             'not-normalize-space','exclude-comments','ext-ges','ext-pes'
             'encoding=', 'xupdate',
             'ezs', # DEPRECATED
             'help', 'verbose', 'version', 'profile=']
    # process command line options
    try:
        (opt, args) = getopt.getopt(args[1:], s_opt, l_opt)
    except getopt.error:
        sys.stderr.write ('Unkwnown option')
        usage(pgm)
        sys.exit(-1)
    recursive, html = 0, 0
    xupd, ezs, verbose = 0, 0, 0
    norm_sp, include_comment, ext_ges, ext_pes = 1, 1, 0, 0
    encoding = 'UTF-8'
    prof = ''
    for o in opt:
        if o[0] == '-r' or o[0] == '--recursive':
            recursive = 1
        elif o[0] == '-H' or o[0] == '--html':
            html = 1
        elif o[0] == '-n' or o[0] == '--not-normalize-space':
            norm_sp = 0
        elif o[0] == '-c' or o[0] == '--exclude-comments':
            include_comment = 0
        elif o[0] == '-g' or o[0] == '--ext-ges':
            ext_ges = 1
        elif o[0] == '-p' or o[0] == '--ext-pes':
            ext_pes = 1
        elif o[0] == '-e' or o[0] == '--encoding':
            encoding = o[1] 
        elif o[0] == '-x' or o[0] == '--xupdate':
            xupd = 1
        elif o[0] == '-z' or o[0] == '--ezs':
            ezs = 1
        elif o[0] == '-v' or o[0] == '--verbose':
            verbose = 1
        elif o[0] == '-p' or o[0] == '--profile':
            prof = o[1] 
        elif o[0] == '-h' or o[0] == '--help':
            usage(pgm)
            sys.exit(0)
        elif o[0] == '-V' or o[0] == '--version':
            from __pkginfo__ import modname, version
            print '%s version %s' % (modname, version)
            sys.exit(0)
    if len(args) != 2:
        usage(pgm)
        sys.exit(-2)
    fpath1, fpath2 = args[0], args[1]
    exit_status = 0
    # if args are directory    
    if os.path.isdir(fpath1) and os.path.isdir(fpath2):
        from logilab.xmldiff.misc import process_dirs, list_print
        common, deleted, added = process_dirs(fpath1, fpath2, recursive)
        
        list_print(deleted[0], 'FILE:', 'deleted')
        list_print(deleted[1], 'DIRECTORY:', 'deleted')
        list_print(added[0], 'FILE:', 'added')
        list_print(added[1], 'DIRECTORY:', 'added')
        exit_status += sum((len(deleted[0]), len(deleted[1]),
                            len(added[0]), len(added[1])))
        for filename in common[0]:
            print '-'*80
            print 'FILE:', filename
            diffs = process_files(os.path.join(fpath1, filename),
                                  os.path.join(fpath2, filename),
                                  norm_sp, xupd, ezs, verbose,
                                  ext_ges, ext_pes, include_comment,
                                  encoding, html)
            if diffs:
                exit_status += diffs
    # if  args are files
    elif os.path.isfile(fpath1) and os.path.isfile(fpath2):
        if prof:
            import profile, pstats, time
            t = time.clock()
            profile.run('process_files(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)'% (
              repr(fpath1), repr(fpath2), repr(norm_sp), repr(xupd),
              repr(ezs), repr(verbose), repr(ext_ges), repr(ext_pes),
              repr(include_comment),repr(encoding), repr(html)),
                        prof)
            print 'Time:',`time.clock()-t`
            p = pstats.Stats(prof)
            p.sort_stats('time','calls').print_stats(.25)
            p.sort_stats('cum','calls').print_stats(.25)
            
        else:
            exit_status = process_files(fpath1, fpath2,
                                        norm_sp, xupd, ezs, verbose,
                                        ext_ges, ext_pes, include_comment,
                                        encoding, html)
    else:
        exit_status = -1
        print fpath1, 'and', fpath2, \
              'are not comparable, or not directory nor regular files'
    sys.exit(exit_status)
    
if __name__ == '__main__':
    run(*sys.argv)



# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" unit tests for xmldiff.
"""
      
import unittest
import cStringIO 
from copy import deepcopy    
#import logilab.xmldiff.ezs
from logilab.xmldiff.objects import *
from logilab.xmldiff.ezs import trees_equal
import sys

class Tree2TreeKnownValues(unittest.TestCase):
    """
    This class check that source_vector functions give known result
    with known input
    """
    
    DATA = cStringIO.StringIO("""
  <memory>
  <mailbox path="/var/spool/mail/syt"/>
  <email_addr mine="yes">syt@logilab.org</email_addr>
  <server-socket1 port="7776" recipe="pia.PDA"/>
   <bookmark-file type="netscape">/home/syt/.netscape/bookmarks.html</bookmark-file>
   <!-- proxy -->
  <server-socket2 port="7777" recipe="proxy.Web proxy"/>
  <junkbuster-method value="18" />
  <spoken-languages>
   <language1 name="english" code="en" />
   <language2 name="francais" code="fr" />
  </spoken-languages>
</memory>
""")   
    from logilab.xmldiff.parser import SaxHandler
    from xml.sax import make_parser
    dh = SaxHandler(1, 1)
    parser = make_parser()
    parser.setContentHandler(dh)
    parser.parse(DATA)
    xmltree1 = dh.get_tree() 

    DATA = cStringIO.StringIO("""
   <memory>
  <mailbox path="/var/spooool/mail/syt"/>
  <email_addr mine="yes">syt@logab.org</email_addr>
  <server-socket1 port="7776" recipe="pia.PDA"/>
  <bookmark-file>/home/syt/.netscape/bookmarks.html</bookmark-file>
  <!-- proxy -->
  <junkbuster-method value="18" />
  <server-socket2 port="7777" recipe="proxy.Web proxy"/>
  <spoken-languages>
   <language1 code="en" name="english"/>
   <language2 name="francais" code="fr" />
   <language3 name="martien" code="ma"/>
  </spoken-languages>
</memory>
""")
    dh.__init__(1, 1)
    parser.parse(DATA)
    xmltree2 = dh.get_tree() 

    # those variables may change if the "DATA" string change
    # this is the post order numbers
    HKNOWN_VALUES = {
        'N/a[0]': 14,
        'a/a[0]': 1,
        'N/a[0]/b[0]': 3,
        'b/a[0]/b[0]': 2,
        'N/a[0]/c[0]': 5,
        'c/a[0]/c[0]': 4,
        'N/a[0]/d[0]': 13,
        'd/a[0]/d[0]': 6,
        'N/a[0]/d[0]/e[0]': 10,
        'e/a[0]/d[0]/e[0]': 7,
        'N/a[0]/d[0]/e[0]/h[0]': 9,
        'h/a[0]/d[0]/e[0]/h[0]': 8,
        'N/a[0]/d[0]/f[0]': 12,
        'f/a[0]/d[0]/f[0]': 11
        }

    dh.__init__(1, 1)
    parser.parse(cStringIO.StringIO("""
    <a>
      <b/>
      <c>
      </c>
      <d>
        <e>
          <h/>
        </e>
        <f>
        </f>
      </d>
    </a>
    """))
    tree1 = dh.get_tree()
                 
    dh.__init__(1, 1)
    parser.parse(cStringIO.StringIO("""
    <a>
      <c/>
      <b>
      </b>
      <d>
        <e>
          <h/>
        </e>
        <f><j/>
        </f>
      </d>
    </a>
    """))
    tree2 = dh.get_tree()


    def setUp (self):
        """ called before each test from this class """
        self.nl1, self.nl2 = [], []

        
##     def test_po_known_values(self):
##         """
##         post_order should give known result with known input
##         """
##         ezsc = ezs.EzsCorrector() 
##         ezsc._post_order(self.tree1, self.nl1, TRUE)
##         num = 1
##         for node in self.nl1[:-1]:
##             self.assert_(num == self.HKNOWN_VALUES[node[N_VALUE] + f_xpath(node)],
##                          'Post order Error for '+f_xpath(node))
##             num = num + 1

##     def test_st_known_values(self):
##         """
##         swap_trees should give known result with known input
##         """
##         ezsc = ezs.EzsCorrector()
##         ezsc._post_order(self.tree1, self.nl1, TRUE)
##         n1 = self.tree1[N_CHILDS][0][N_CHILDS][1]
##         s1 = self.tree1[N_CHILDS][0][N_CHILDS][2]
##         self.assert_(ezs.swap_trees(n1, s1, n1, s1) == ezs.C_INFINI, 'SWAP')
##         self.assert_(ezs.swap_trees(n1, s1, s1, n1) == ezs.C_SWAP, 'SWAP')

##     def test_ezs_known_values(self):
##         """
##         ezs should give known result with known input
##         """
##         ezsc = ezs.EzsCorrector()
##         actions = ezsc.process_trees(self.tree1, self.tree2)
##         action_list1= ['append', 'swap']
##         for i in range(len(actions)):
##             self.assert_(actions[i][A_DESC] == action_list1[i])

##         actions = ezsc.process_trees(self.xmltree1, self.xmltree2)
##         #import format
##         #format.factions_print(actions)
##         action_list1= ['insert-after', 'remove', 'update', 'update', 'swap']
##         for i in range(len(actions)):
##             self.assert_(actions[i][0] == action_list1[i])

        
    def test_sanity(self):
        """
        global sanity check for do() function
        """
        self.assert_(trees_equal(self.tree1, self.tree1),
                     'Tree unequal error')
        self.assert_(not trees_equal(self.tree1, self.tree2),
                     'Tree equal error')
        #        print  "SOURCE TREE"
        #        xml_print(self.tree1)
        #        print  "DESTINATION TREE"
        #        xml_print(self.tree2)


def suite():
    """return the unitest suite"""
    loader = unittest.TestLoader()
    module = sys.modules[__name__]
    if __name__ == '__main__' and len(sys.argv) > 1:
        return loader.loadTestsFromNames(sys.argv[1:], module)
    return loader.loadTestsFromModule(module)

def Run(runner=None):
    """run tests"""
    testsuite = suite()
    if runner is None:
        runner = unittest.TextTestRunner()
        # uncomment next line to write tests results in a file
        #runner.__init__(open('tests.log','w+'))    
    return runner.run(testsuite)
    
if __name__ == '__main__':
    Run()



# -*- coding: ISO-8859-1 -*-
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2002-2004 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr  
"""

__revision__ = '$Id: __pkginfo__.py,v 1.9 2004/12/22 16:10:14 alf Exp $'

modname = 'xmldiff'

numversion = (0, 6, 6)
version = '.'.join(map(str, numversion))

license = 'GPL'
copyright = '''Copyright © 2001-2004 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr'''

short_desc = "tree 2 tree correction between xml documents"
long_desc = """Xmldiff is a utility for extracting differences between two
xml files. It returns a set of primitives to apply on source tree to obtain
the destination tree.
.
The implementation is based on _Change detection in hierarchically structured 
- information_, by S. Chawathe, A. Rajaraman, H. Garcia-Molina and J. Widom, 
- Stanford University, 1996"""

author = "Sylvain Th\xc3\xa9nault"
author_email = "sylvain.thenault@logilab.fr"

web = "http://www.logilab.org/projects/%s" % modname
ftp = "ftp://ftp.logilab.org/pub/%s" % modname
mailinglist = 'xml-projects@logilab.org'

subpackage_of = 'logilab'


from os.path import join

scripts = [join('bin', 'xmldiff'), join('bin', 'xmlrev')]
include_dirs = [join('test', 'data')]

try:
    from distutils.core import Extension
    ext_modules = [Extension('logilab.xmldiff.maplookup',
                             ['extensions/maplookup.c'])]
except:
    pass

data_files = [("share/sgml/stylesheet/xmldiff",
               ['xsl/docbook_rev.xsl', 'xsl/xmlrev.xslt'])
              ]



"""
longest common subsequence algorithm

the algorithm is describe in "An O(ND) Difference Algorithm and its Variation"
by Eugene W. MYERS

As opposed to the algorithm in difflib.py, this one doesn't require hashable
elements 
"""
__revision__ = '$Id: mydifflib.py,v 1.8 2004/09/03 10:06:39 alf Exp $'

def lcs2(X, Y, equal):
    """
    apply the greedy lcs/ses algorithm between X and Y sequence
    (should be any Python's sequence)
    equal is a function to compare X and Y which must return 0 if
    X and Y are different, 1 if they are identical
    return a list of matched pairs in tuplesthe greedy lcs/ses algorithm
    """
    N, M = len(X), len(Y)
    if not X or not Y :
        return []
    max = N + M
    v = [0 for i in xrange(2*max+1)]
    common = [[] for i in xrange(2*max+1)]
    for i in xrange(max+1):
        for j in xrange(-i, i+1, 2):
            if j == -i or j != i and v[j-1] < v[j+1]:
                x = v[j+1]
                common[j] = common[j+1][:]
            else:
                x = v[j-1] + 1
                common[j] = common[j-1][:]
                
            y = x - j
            while x < N and y < M and equal(X[x], Y[y]):
                common[j].append((X[x], Y[y]))
                x += 1 ; y += 1

            v[j] = x
            if x >= N and y >= M:
                return common[j]

def lcsl(X, Y, equal):
    """return the length of the result sent by lcs2"""
    return len(lcs2(X,Y,equal))

def quick_ratio(a,b):
    """
    optimized version of the standard difflib.py quick_ration
    (without junk and class)
    Return an upper bound on ratio() relatively quickly.
    """
    # viewing a and b as multisets, set matches to the cardinality
    # of their intersection; this counts the number of matches
    # without regard to order, so is clearly an upper bound
    if not a and not b:
        return 1
    fullbcount = {}
    for elt in b:
        fullbcount[elt] = fullbcount.get(elt, 0) + 1
    # avail[x] is the number of times x appears in 'b' less the
    # number of times we've seen it in 'a' so far ... kinda
    avail = {}
    availhas, matches = avail.has_key, 0
    for elt in a:
        if availhas(elt):
            numb = avail[elt]
        else:
            numb = fullbcount.get(elt, 0)
        avail[elt] = numb - 1
        if numb > 0:
            matches = matches + 1
    return 2.0 * matches / (len(a) + len(b))

try:
    import os
    if os.environ.get('PYLINT_IMPORT') != '1': # avoid erros with pylint
        import psyco
        psyco.bind(lcs2)
except Exception, e:
    pass

def test():
    """
    FIXME this should go into the test suite.
    """
    import time
    t = time.clock()
    quick_ratio('abcdefghijklmnopqrst'*100, 'abcdefghijklmnopqrst'*100)
    print 'quick ratio :',time.clock()-t
    lcs2('abcdefghijklmnopqrst'*100, 'abcdefghijklmnopqrst'*100,
         lambda x, y : x==y)
    print 'lcs2 :       ',time.clock()-t
    quick_ratio('abcdefghijklmno'*100, 'zyxwvutsrqp'*100)
    print 'quick ratio :',time.clock()-t
    lcs2('abcdefghijklmno'*100, 'zyxwvutsrqp'*100, lambda x, y : x==y)
    print 'lcs2 :       ',time.clock()-t
    quick_ratio('abcdefghijklmnopqrst'*100, 'abcdefghijklmnopqrst'*100)
    print 'quick ratio :',time.clock()-t
    lcs2('abcdefghijklmnopqrst'*100, 'abcdefghijklmnopqrst'*100,
         lambda x, y : x==y)
    print 'lcs2 :       ',time.clock()-t
    quick_ratio('abcdefghijklmno'*100, 'zyxwvutsrqp'*100)
    print 'quick ratio :',time.clock()-t
    lcs2('abcdefghijklmno'*100, 'zyxwvutsrqp'*100, lambda x, y : x==y)
    print 'lcs2 :       ',time.clock()-t

    
if __name__ == '__main__':
    print lcsl('abcde', 'bydc', lambda x, y : x==y)
    for a in lcs2('abcde', 'bydc', lambda x, y : x==y):
        print a
    print lcsl('abacdge', 'bcdg', lambda x, y : x==y)
    for a in lcs2('abacdge', 'bcdg', lambda x, y : x==y):
        print a
        
